#!/opt/local/bin/perl -w
#
#    make_aors [-f]
#
# Generate KINGFISH AORS from input file in inputs/, assigning
# observational details based on the "package" for three different
# surface brightness bins.  If -f is passed, strips are grouped with
# nuclei into single "FullStrip" strips, saving some time.
#
# JD Smith, Feb, 2010
#
use Getopt::Std;
getopts('fc');

use Data::Dumper;
use Clone qw(clone);

my $pacs_fov=47.;  # PACS Field of view, Arcsec
my $ply=2       ;  # Portion of PACS field to step, parallel to strip
my $dither=4.5  ;  # Dither size (cross-strip, and nuc/enucs) (0.5 increments)
my $unit_limit=600; #Max units in concatenation
my $do_line_dither=0;


# AOR packages (repitions/cycles, not including the raster ply),
# keyed to corrected TIR SB, in three bins

@tir_bins=(90,225); # Cuts in un-dilution corrected TIR SB


$package{nuc}=
  {"faint" =>  {max=>$tir_bins[0],
		aors=>[{order=>2,cycles=>2,
			line_reps=>{'cII'=>2,'oIII'=>3}},
		       {order=>3,cycles=>3,
			line_reps=>{'nII 122'=>5,'oI'=>5}}]},

   "medium" => {min=>$tir_bins[0], max=> $tir_bins[1],
		aors=>[{order=>3,cycles=>1,
			line_reps=>{'cII'=>2,'oI'=>3}},
		       {order=>2,cycles=>2,
			line_reps=>{'nII 122'=>3,'oIII'=>2}}]},

   "bright" => {min=>$tir_bins[1],
		aors=>[{order=>3,cycles=>1,
			line_reps=>{'cII'=>1,'oI'=>2,'nII 122'=>3}},
		       {order=>2,cycles=>2,
			line_reps=>{'nII 205'=>4,'oIII'=>1}}]}};


# Remove [NII] from faint enuc and strips
$package{enuc}=clone($package{nuc});
delete $package{enuc}{faint}{aors}[1]{line_reps}{'nII 122'};

$package{strip}=clone($package{enuc});


# For fullstrip's, pick up the missing [NII] lines in the nucleus
$package{faintstripnuc}=clone($package{nuc});
foreach (keys %{$package{faintstripnuc}}) {
  shift @{$package{faintstripnuc}{$_}{aors}};
};


delete $package{faintstripnuc}{faint}{aors}[0]{line_reps}{oI};
delete $package{faintstripnuc}{medium}{aors}[0]{line_reps}{oIII};
delete $package{faintstripnuc}{bright}{aors}[0]{line_reps}{oIII};
$package{faintstripnuc}{bright}{aors}[0]{line_reps}{'nII 122'}=2;


$package{mediumstripnuc}{bright}=clone($package{nuc}{bright});
shift @{$package{mediumstripnuc}{bright}{aors}};
delete $package{mediumstripnuc}{bright}{aors}[0]{line_reps}{oIII};


#print Dumper(%package),"\n";



# Just print cycles
if ($opt_c) {
  foreach $bin (qw(nuc enuc strip)) {
    printf "==> $bin\n";
    foreach $level (qw(bright medium faint)) {
      printf "%6s: ",$level;
      print join ("; ",
		  map {
		    my $lr=$_->{line_reps};
		    "$_->{cycles} x [" .
		      join(", ",map {"$_: $lr->{$_}"} sort keys %$lr) . "]";
		  } @{$package{$bin}{$level}{aors}}),"\n";
    }
  }
  exit $opt_c;
}



%lines=("oI"   =>
	{id => "O I 3P1-3P2",  wave => 63.18,  f_FIR=>2.e-3},
	"oIII" =>
	{id => "O III 3P1-3P0",wave => 88.36,  f_FIR=>1.e-3},
	"nII 122" =>
	{id => "N II 3P2-3P1", wave => 121.9,  f_FIR=>3.e-4},
	"cII"  =>
	{id => "CII C+",       wave => 157.741,f_FIR=>3.e-3},
	"nII 205" =>
	{id => "N II 3P1-3P0", wave => 205.18, f_FIR=>2.e-4});

sub hex_to_float {
  my @ret=map{
    my @parts=split(/:/);
    if ($#parts==0) { # No colons
      @parts=/([-+]?[0-9]{2}(?:\.[0-9]+)?)/g # E.g. 081808.2
    }
    my $sign;
    if ($parts[0]=~/^[-+]/) {
      $sign=substr($parts[0],0,1);
      $parts[0]=substr($parts[0],1);
    }
    my $val=$parts[0]+$parts[1]/60.+$parts[2]/60./60.;
    $val=-$val if $sign && $sign eq "-";
    $val;
  } @_;
  $#ret==0?$ret[0]:@ret;
}

sub deg_to_hms {
  my ($ra,$dec)=@_;
  if($ra=~/[0-9]:/) {  # colon separated to [hd]ms, assume in hours already
    my $c=0;
    $ra=join("",map{$_.qw(h m s)[$c++]} split(/:/,$ra));
    $c=0;
    $dec=join("",map{$_.qw(d m s)[$c++]} split(/:/,$dec));
  } else {
    # decimal degrees in both cases
    $ra/=15.;
    my $h=int($ra);
    my $m=int(60.*($ra-$h));
    my $s=60.*(60.*($ra-$h)-$m);
    if (sprintf("%05.2f",$s) eq "60.00") {
      $s=0.0; # avoid rounding to 60.0s
      $m++;
      if ($m==60) {
    	$m=0;
    	$h++;
      }
    }
    $ra=sprintf("%02dh%02dm%05.2fs",$h,$m,$s);

    my $d=int($dec);
    $m=int(60.*(abs($dec)-abs($d)));
    $s=60.*(60.*(abs($dec)-abs($d))-$m);
    if (sprintf("%04.1f",$s) eq "60.0") {
      $s=0.0;  # avoid rounding to 60.0s
      $m++;
      if ($m==60) {
	$m=0;
	$d++;
      }
    }
    $dec=sprintf("%1s%02dd%02dm%04.1fs",$dec<0.?"-":"+",abs($d),$m,$s);
  }
  return ($ra,$dec);
}



# Compute the strip center, either separately for two sub-strips, or
# for a single combined strip (including the nucleus).
sub strip_center {
  my ($rec,$full)=@_;
  my ($ra,$dec)=hex_to_float(@{$rec}{qw(ra dec)});
  $ra*=15.;
  my $pa=$rec->{pa};
  my @n=@{$rec->{nstrip}};
  #print "\nN: ",join(",",@n),"  RA: $ra, DEC: $dec\n";
  my $coords;
  my $radeg=180./3.14159265358979;
  # Step dimension, degrees decimal
  my @d=(sin($pa/$radeg)/cos($dec/$radeg)*$pacs_fov/3600.,
	 cos($pa/$radeg)*$pacs_fov/3600.);
  if($full) {
    my $c_ra=$ra  + ($n[1]-$n[0])/2.*$d[0];
    my $c_dec=$dec+ ($n[1]-$n[0])/2.*$d[1];
    $coords=[$c_ra,$c_dec];
  } else {
    my $c_ra_plus =$ra   + ($n[1]+1)/2.*$d[0];
    my $c_ra_minus=$ra   - ($n[0]+1)/2.*$d[0];
    my $c_dec_plus=$dec  + ($n[1]+1)/2.*$d[1];
    my $c_dec_minus=$dec - ($n[0]+1)/2.*$d[1];
    $coords=[$c_ra_minus,$c_dec_minus,$c_ra_plus,$c_dec_plus];
  }

  return $coords;
}

# Compute the number of strip parallel steps required to cover a given
# number of "pacs fields" (i.e. 47") at full coverage.  MAX_PLY is the
# maximum number of visit a given position gets in steps along the
# strip.  E.g. a step of 47"/2 would give 2-ply, etc.  The larger the
# "ply", the more less-than-full-depth coverage you have at the ends
# of the strips.
sub strip_parallel_steps {
  my ($n_pacs_fields)=@_;
  return $n_pacs_fields*$ply + ($ply-1);
}

sub line_dither {
  return .05 # FIXME, amount in um for dithering line in multi-reps
}

###############################################################################

# Read Sample
open FILE, "<inputs/kingfish_sample_final.txt";
while (<FILE>) {
  next if $.<=5;
  last if /^-{10,}/;
  chomp;
  my ($gal,$ra,$dec,$cz)=
    /([A-Za-z0-9]+)[ \t]+([0-9:.]+)[ \t]+([+-][0-9:.]+).*?[0-9.]x[0-9.]+[ \t]+([-0-9]+)/;/;/;
  $gal=lc $gal;
  @{$obs{$gal}}{qw(ra dec cz)}=($ra,$dec,int($cz));
}
close FILE;

# Read TIR/enuc/strip data
open FILE, "<kingfish_enuc_aor.txt";
while (<FILE>) {
  next if $.<=2;
  chomp;
  my ($name,$ra,$dec,$tir)=split(/ +/);
  my ($gal,$enuc)=($1,$2) if $name=~/([a-z0-9]+)_([a-z0-9]+)/;
  @{$obs{$gal}{enuc}{$enuc}}{qw(ra dec tir)}=($ra,$dec,$tir);
}
close FILE;

open FILE, "<kingfish_nuc_aor.txt";
while (<FILE>) {
  next if $.<=2;
  chomp;
  my ($gal,$ra,$dec,$tir)=split(/ +/);
  @{$obs{$gal}{nuc}}{qw(ra dec tir)}=($ra,$dec,$tir);
}
close FILE;

open FILE, "<kingfish_strip_aor.txt";
while (<FILE>) {
  next if $.<=2;
  chomp;
  my ($name,$ra,$dec,$tir,$pa,$nplus,$nminus)=split(/ +/);
  my $gal=$1 if $name=~/([a-z0-9]+)_strip/;
  @{$obs{$gal}{strip}}{qw(ra dec pa nstrip tir)}=
    ($ra,$dec,$pa,[$nplus,$nminus],$tir);
}
close FILE;

# Read Offset Positions
open FILE, "<inputs/kingfish_pacs_offs.txt";
while (<FILE>) {
  chomp;
  my ($gal,$ra,$dec)=split(/ +/);
  $obs{lc $gal}{offset}=[sprintf("%0.9f",hex_to_float($ra)*15.),
			 sprintf("%0.9f",hex_to_float($dec))];
}
close FILE;


#print Dumper(%obs),"\n";
#__END__

sub print_aor_header {
  print <<EOF;
<?xml version="1.0" encoding="UTF-8"?>
  <v07:RequestHolder xmlns:v07="http://www.esa.int/herschel/aors/v07">
    <v07:AorIoVersion>0.7</v07:AorIoVersion>
EOF
}

# Print an AOR of title $title, for galaxy $gal
# given a record with fields:
#    @{$record}{offset,cz}
# Along with observation parameters with fields (a hash)
#    @{$obs}{ra,dec,x_step,y_step,pa,nx,ny,tir}
# and a reference to a hash of AOR details::
#    @{$aor}{cycles,order,line_reps=>{line_name=>repitition,...}}
# where line_name is a field of the global %lines hash.
sub print_aor {
  my ($title,$gal,$record,$mode,$aor)=@_;

  my $full_title=$title." - ".join(", ",sort(keys %{$aor->{line_reps}}));


  print <<EOF;
   <v07:Request>
      <v07:Title>$full_title</v07:Title>
      <v07:Status>new</v07:Status>
      <v07:Comment xsi:nil="true" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"/>
      <v07:ModeType>
         <v07:PacsLineSpec>
            <v07:order>order$aor->{order}</v07:order>
            <v07:source>large</v07:source>
            <v07:Mapping>
               <v07:PacsRaster>
                  <v07:x>$mode->{x_step}</v07:x>
                  <v07:y>$mode->{y_step}</v07:y>
                  <v07:angle>$mode->{pa}</v07:angle>
                  <v07:m>$mode->{nx}</v07:m>
                  <v07:n>$mode->{ny}</v07:n>
                  <v07:mapRasterAngleRefFrame>true</v07:mapRasterAngleRefFrame>
               </v07:PacsRaster>
               <v07:mapRefParams>
                  <v07:offsetSel>false</v07:offsetSel>
                  <v07:OffsetParams>
                     <v07:Position>
                        <v07:PositionRa>$record->{offset}[0]</v07:PositionRa>
                        <v07:PositionDec>$record->{offset}[1]</v07:PositionDec>
                     </v07:Position>
                  </v07:OffsetParams>
               </v07:mapRefParams>
            </v07:Mapping>
            <v07:chopNod>false</v07:chopNod>
            <v07:faintLines>true</v07:faintLines>
            <v07:chopAvoidFrom>0.0</v07:chopAvoidFrom>
            <v07:chopAvoidTo>0.0</v07:chopAvoidTo>
            <v07:throw>large</v07:throw>
            <v07:redshiftType>optical</v07:redshiftType>
            <v07:redshiftValue>$record->{cz}</v07:redshiftValue>
            <v07:repetition>$aor->{cycles}</v07:repetition>
EOF

  my $units=0;
  foreach my $line_name (sort keys %{$aor->{line_reps}}) {
    my $line_wave=$lines{$line_name}{wave}*(1.+$record->{cz}/299792.46);
    my $line_id=$lines{$line_name}{id};
    my $line_flux=$lines{$line_name}{f_FIR}/2.*$mode->{tir}*1.e-7*
      9.4*9.4/206265.**2*1.e18; # Flux in 1.e-18 W/m^2
    my $line_dither=line_dither($line_wave);
    my $reps=$aor->{line_reps}{$line_name};

    # Collect total number of units in this AOR
    $units+=$reps*$aor->{cycles}*$mode->{nx}*$mode->{ny};

    if($do_line_dither) {
      for (my $r=0;$r<$reps;$r++) {
	# Multiple nudged lines
	my $shift_wavelength=$line_wave + ($r-int($reps/2))*$line_dither;

	print <<EOF;
            <v07:line>
               <v07:wavelengthId>$line_id</v07:wavelengthId>
               <v07:wavelength>$shift_wavelength</v07:wavelength>
               <v07:fluxUnits>fluxWatt</v07:fluxUnits>
               <v07:lineFlux>$line_flux</v07:lineFlux>
               <v07:contFlux>0.0</v07:contFlux>
               <v07:lineWidth>1.0</v07:lineWidth>
               <v07:widthUnits>kms</v07:widthUnits>
               <v07:lineRepetition>1</v07:lineRepetition>
            </v07:line>
EOF
      }
    } else {
      print <<EOF;
            <v07:line>
               <v07:wavelengthId>$line_id</v07:wavelengthId>
               <v07:wavelength>$line_wave</v07:wavelength>
               <v07:fluxUnits>fluxWatt</v07:fluxUnits>
               <v07:lineFlux>$line_flux</v07:lineFlux>
               <v07:contFlux>0.0</v07:contFlux>
               <v07:lineWidth>1.0</v07:lineWidth>
               <v07:widthUnits>kms</v07:widthUnits>
               <v07:lineRepetition>$reps</v07:lineRepetition>
            </v07:line>
EOF
    }
  }
  #print "MODE: ",Dumper($mode),"\n";
  my ($ra,$dec)=deg_to_hms(@{$mode}{qw(ra dec)}); 

  print <<EOF;
         </v07:PacsLineSpec>
      </v07:ModeType>
      <v07:TargetType>
         <v07:FixedSingle>
            <v01:Name xmlns:v01="http://www.esa.int/herschel/targets/v01">$gal</v01:Name>
            <v01:AvoidEarth xmlns:v01="http://www.esa.int/herschel/targets/v01">true</v01:AvoidEarth>
            <v01:AvoidOthers xmlns:v01="http://www.esa.int/herschel/targets/v01">true</v01:AvoidOthers>
            <v01:CoordSystem xmlns:v01="http://www.esa.int/herschel/targets/v01">Equatorial</v01:CoordSystem>
            <v01:Equinox xmlns:v01="http://www.esa.int/herschel/targets/v01">J2000</v01:Equinox>
            <v01:RaLon xmlns:v01="http://www.esa.int/herschel/targets/v01">$ra</v01:RaLon>
            <v01:DecLat xmlns:v01="http://www.esa.int/herschel/targets/v01">$dec</v01:DecLat>
            <v01:PropMo xmlns:v01="http://www.esa.int/herschel/targets/v01">
               <v01:Lon>0.0</v01:Lon>
               <v01:Lat>0.0</v01:Lat>
            </v01:PropMo>
            <v01:Epoch xmlns:v01="http://www.esa.int/herschel/targets/v01">2000.0</v01:Epoch>
         </v07:FixedSingle>
      </v07:TargetType>
   </v07:Request>
EOF

  return [$full_title,$units];
}

sub print_concatenate_constraints {
  my ($constraints)=@_;
  print <<EOF;
   <v07:Constraints>
EOF

  foreach $gal (sort keys %$constraints) {
    my %concat;
    my ($ns,$en)=@{$constraints->{$gal}}{qw(nucstrip enuc)};

    my ($nscnt,$encnt)=(0,0);
    foreach $aor (@$ns) {$nscnt+=$aor->[1];}
    foreach $aor (@$en) {$encnt+=$aor->[1];}
    if($nscnt+$encnt<$unit_limit) {
      $concat{$gal}=[map {$_->[0]} @$ns,@$en];
    } else {
      warn "$gal concatenation exceeded limit, splitting Strips/enucs\n";
      my $name=$opt_f?"FullStrip":"Nucleus/Strips";
      $concat{"$gal $name"}=[map {$_->[0]} @$ns];
      $concat{"$gal enucs"}=[map {$_->[0]} @$en] if $encnt>0;
    }


    foreach $name (sort keys %concat) {
    
      print <<EOF;
      <v07:Concatenation>
         <v01:Comment xmlns:v01="http://www.esa.int/herschel/constraints/v01"/>
         <v01:Name xmlns:v01="http://www.esa.int/herschel/constraints/v01">Concatenate $name</v01:Name>
EOF

      foreach $title (@{$concat{$name}}) {
	print <<EOF;
	   	<v01:Request xmlns:v01="http://www.esa.int/herschel/constraints/v01">$title</v01:Request>
EOF
      }
      print <<EOF;
      </v07:Concatenation>
EOF
    }
  }
  print <<EOF;
   </v07:Constraints>
EOF
}

sub print_aor_footer {
    print <<EOF;
</v07:RequestHolder>
EOF
}

sub tir_bin {
  my ($type,$rec)=@_;
  my ($bin,$level);
  foreach $bin_name (keys %{$package{$type}}) {
    my ($min,$max)=@{$package{$type}{$bin_name}}{qw(min max)};
    if ((!defined($min) || $rec->{tir}>$min) &&
	(!defined($max) || $rec->{tir}<$max)) {
      # This is the bin, use it
      $level=$bin_name;
      $bin=$package{$type}{$level} if defined($package{$type}{$level});
      last;
    }
  }
  return ($level,$bin);
}
  
  
# Select a bin and print a set of AORs for this record
sub do_aor {
  my ($gal,$type,$rec,$title,$n,$xstep,$concat)=@_;

  # Find the brightness bin
  my ($level,$bin)=tir_bin($type,$rec);

  my $pa=defined($rec->{pa})?$rec->{pa}:0.0;
  # Print the corresponding AORs
  foreach $aor (@{$bin->{aors}}) {
    #$aor => [{order=>, cycles=>,line_reps=>{name=>reps,...}}
    push @$concat,
      print_aor("$gal $title [$level]", # Title
		$gal,       # Gal
		$obs{$gal}, # Record
		{ra=>$rec->{ra},  # Mode
		 dec=>$rec->{dec},
		 x_step=>$xstep,y_step=>$dither,,
		 pa=>$pa,nx=>$n->[0],ny=>$n->[1],tir=>$rec->{tir}},
		$aor);      # AOR
  };
}



##############################################################################
my $concat;

print_aor_header();
foreach $gal (sort keys %obs) {
  next unless exists $obs{$gal}{nuc};

  next if $gal eq 'ngc1097' || $gal eq 'ngc4559';
  #next unless $gal=~/^ngc54[05]/;

  my (@concat,@enucconcat);

  # Regular Nucleus (only if not using FullStrip, or if no strip exists)
  unless($opt_f && defined($obs{$gal}{strip})) {
    do_aor($gal,"nuc",$obs{$gal}{nuc},"Nucleus",[2,2],$dither,\@concat);
  }

  #--- FullStrip "extra" nucleus
  if($opt_f && exists($obs{$gal}{strip})) {
    # FullStrip will be used, check to see if we need separate nuc aors
    my ($stripbin)=tir_bin("strip",$obs{$gal}{strip});
    if ($stripbin ne "bright") {
      do_aor($gal,$stripbin."stripnuc",$obs{$gal}{nuc},"Nucleus Extra",[2,2],
	     $dither,defined($obs{$gal}{enuc})?\@enucconcat:\@concat);
    }
  }

  #-- extra-nuclear
  foreach $enuc (sort keys %{$obs{$gal}{enuc}}) {
    do_aor($gal,"enuc",$obs{$gal}{enuc}{$enuc},"Enuc_$enuc",[2,2],$dither,
	   \@enucconcat);
  }

  #-- strips
  if(defined($obs{$gal}{strip})) {
    my $strip_center=strip_center($obs{$gal}{strip},$opt_f);
    if ($opt_f) {
      # Full strips incorporating nucleus
      if ($obs{$gal}{nuc}{tir}<$obs{$gal}{strip}{tir}) {
	warn("$gal: nucleus fainter than strip: $obs{$gal}{nuc}{tir} < $obs{$gal}{strip}{tir}.  Combining anyway.\n");
      }
      # Group strips and nuclei into "super strips"
      @{$obs{$gal}{strip}}{qw(ra dec)}=@{$strip_center};
      my $steps=strip_parallel_steps($obs{$gal}{strip}{nstrip}[0]+
				     $obs{$gal}{strip}{nstrip}[1]+1);
      do_aor($gal,"strip",$obs{$gal}{strip},"FullStrip",[$steps,2],
	     $pacs_fov/$ply,\@concat);
    } else {
      # Two separate strips
      for ($s=0;$s<=1;$s++) {
	my $n=$obs{$gal}{strip}{nstrip}[$s];

	# Change the RA to the appropriate side's center
	@{$obs{$gal}{strip}}{qw(ra dec)}=@{$strip_center}[2*$s..2*$s+1];
	my $steps=strip_parallel_steps($n);
	#print "$gal strip $n -> $steps\n";
	# Change the RA to the appropriate strip center	
	do_aor($gal,"strip",$obs{$gal}{strip},"Strip".($s?"Plus":"Minus"),
	       [$steps,2],$pacs_fov/$ply,\@concat);
      }
    }
  }

  $concat{$gal}{nucstrip}=\@concat;
  $concat{$gal}{enuc}=\@enucconcat;
}

print_concatenate_constraints(\%concat);

print_aor_footer();
